/*
 * Copyright 2009 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of either or both of the following licenses:
 *
 * 1) the GNU Lesser General Public License version 3, as published by the
 * Free Software Foundation; and/or
 * 2) the GNU Lesser General Public License version 2.1, as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the applicable version of the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of both the GNU Lesser General Public
 * License version 3 and version 2.1 along with this program.  If not, see
 * <http://www.gnu.org/licenses/>
 *
 * Authored by: Neil Jagdish Patel <neil.patel@canonical.com>
 *
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "ctk-text.h"

#include <math.h>

G_DEFINE_TYPE (CtkText, ctk_text, CLUTTER_TYPE_TEXT);

#define CTK_TEXT_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj),\
  CTK_TYPE_TEXT, \
  CtkTextPrivate))

#define GFLOAT_FROM_PANGO_UNIT(x) (pango_units_to_double (x))
#define GFLOAT_TO_PANGO_UNIT(x) (pango_units_from_double (x))

struct _CtkTextPrivate
{
  GtkSettings    *settings;
  PangoAlignment  alignment;

  gboolean        font_overridden;
};

/* Globals */

/* Forwards */
static void on_gtk_font_changed          (GtkSettings *settings,
    GParamSpec  *pspec,
    ClutterText *text);
static void on_clutter_text_font_changed (CtkText *self,
    GParamSpec *pspec);

/* GObject stuff */
static void
ctk_text_allocate (ClutterActor          *actor,
                   const ClutterActorBox *box,
                   ClutterAllocationFlags flags)
{
  CtkTextPrivate *priv = CTK_TEXT (actor)->priv;
  ClutterActorBox real_box;
  PangoLayout    *layout;

  layout = clutter_text_get_layout (CLUTTER_TEXT (actor));

  if (priv->alignment != PANGO_ALIGN_LEFT
      && pango_layout_get_line_count (layout) == 1)
    {
      PangoRectangle log_rect;
      gfloat         real_width;
      gfloat         text_width;

      pango_layout_get_extents (layout, NULL, &log_rect);
      text_width = GFLOAT_FROM_PANGO_UNIT (log_rect.width);
      real_width = box->x2 - box->x1;

      if (priv->alignment == PANGO_ALIGN_CENTER)
        {
          real_box.x1 = (real_width/2) - (text_width/2);
          real_box.x2 = real_box.x1 + text_width;
        }
      else //PANGO_ALIGN_RIGHT
        {
          real_box.x1 = box->x2 - text_width;
          real_box.x2 = box->x2;
        }
    }
  else
    {
      real_box.x1 = box->x1;
      real_box.x2 = box->x2;
    }

  real_box.x1 = floor (real_box.x1);
  real_box.x2 = floor (real_box.x2);
  real_box.y1 = floor (box->y1);
  real_box.y2 = floor (box->y2);

  CLUTTER_ACTOR_CLASS (ctk_text_parent_class)->allocate (actor,
      &real_box,
      flags);
}

static void
ctk_text_constructed (GObject *object)
{
  CtkTextPrivate *priv = CTK_TEXT (object)->priv;
  ClutterText    *text = CLUTTER_TEXT (object);
  ClutterColor    white = { 0xff, 0xff, 0xff, 0xff };
  gchar          *font = NULL;

  g_object_get (G_OBJECT (priv->settings), "gtk-font-name", &font, NULL);

  clutter_text_set_font_name (text, font);
  clutter_text_set_color (text, &white);

  g_free (font);
}

static void
ctk_text_focus_in (ClutterActor *actor)
{
  g_debug ("CtkText: focus-in");
}

static void
ctk_text_class_init (CtkTextClass *klass)
{
  GObjectClass      *obj_class = G_OBJECT_CLASS (klass);
  ClutterActorClass *act_class = CLUTTER_ACTOR_CLASS (klass);

  obj_class->constructed  = ctk_text_constructed;
  act_class->allocate     = ctk_text_allocate;
  act_class->key_focus_in = ctk_text_focus_in;

  g_type_class_add_private (obj_class, sizeof (CtkTextPrivate));
}

static void
ctk_text_init (CtkText *text)
{
  CtkTextPrivate *priv;

  priv = text->priv = CTK_TEXT_GET_PRIVATE (text);

  priv->font_overridden = FALSE;
  priv->alignment = PANGO_ALIGN_LEFT;

  priv->settings = gtk_settings_get_default ();
  g_signal_connect (priv->settings, "notify::gtk-font-name",
                    G_CALLBACK (on_gtk_font_changed), text);

  g_signal_connect (text, "notify::font-name",
                    G_CALLBACK (on_clutter_text_font_changed), NULL);
}

ClutterActor *
ctk_text_new (const gchar *text)
{
  return g_object_new (CTK_TYPE_TEXT,
                       "text", text,
                       "use-markup", TRUE,
                       NULL);
}

/*
 * Private methods
 */
static void
on_gtk_font_changed (GtkSettings *settings,
                     GParamSpec  *pspec,
                     ClutterText *self)
{
  gchar *font_name = NULL;

  g_return_if_fail (CTK_IS_TEXT (self));

  /* If the developer has chosen their own font, then we don't auto-update
   * the text font with the system default
   */
  if (CTK_TEXT (self)->priv->font_overridden)
    return;

  g_object_get (G_OBJECT (settings), "gtk-font-name", &font_name, NULL);

  clutter_text_set_font_name (self, font_name);

  g_free (font_name);
}

static void
on_clutter_text_font_changed (CtkText *self,
                              GParamSpec *pspec)
{
  CtkTextPrivate *priv;
  gchar          *sys_font;
  const gchar    *text_font;

  g_return_if_fail (CTK_IS_TEXT (self));
  priv = self->priv;

  g_object_get (G_OBJECT (priv->settings), "gtk-font-name", &sys_font, NULL);
  text_font = clutter_text_get_font_name (CLUTTER_TEXT (self));

  if (g_strcmp0 (sys_font, text_font) == 0)
    priv->font_overridden = FALSE;
  else
    priv->font_overridden = TRUE;

  g_free (sys_font);
}

/*
 * Public methods
 */
void
ctk_text_set_alignment (CtkText        *self,
                        PangoAlignment  alignment)
{
  CtkTextPrivate *priv;

  g_return_if_fail (CTK_IS_TEXT (self));
  priv = self->priv;

  if (priv->alignment == alignment)
    return;

  priv->alignment = alignment;

  if (CLUTTER_ACTOR_IS_VISIBLE (self))
    clutter_actor_queue_relayout (CLUTTER_ACTOR (self));
}

PangoAlignment
ctk_text_get_alignment (CtkText *self)
{
  g_return_val_if_fail (CTK_IS_TEXT (self), PANGO_ALIGN_LEFT);

  return self->priv->alignment;
}

